from pydantic_client.decorators import delete, get, patch, post, put
from pydantic_client.factory import PydanticClientFactory
from pydantic_client.main import PydanticClient
from pydantic_client.schema.client_config import ClientConfig

__all__ = [
    "PydanticClient",
    "PydanticClientFactory",
    "ClientConfig",
    "patch",
    "get",
    "post",
    "put",
    "delete",
]



import inspect
import logging
import re
from typing import Dict, Any
from urllib.parse import parse_qsl, urlparse

from pydantic import BaseModel

from pydantic_client.clients.abstract_client import AbstractClient
from pydantic_client.schema.http_request import HttpRequest
from pydantic_client.schema.method_info import MethodInfo

logger = logging.getLogger("pydantic_client.parser")


class Parser:
    querystring_pattern = re.compile(r"\{(.*?)\}")

    @staticmethod
    def _apply_args(
        method_info: MethodInfo,
        *args, **kwargs
    ) -> Dict[str, Any]:
        f = inspect.getcallargs(
            method_info.func, *args, **kwargs,
        )
        f.pop("self", None)
        for key, value in f.items():
            if isinstance(value, BaseModel):
                f[key] = value.model_dump()
        return f

    @staticmethod
    def _get_url(method_info, args) -> str:
        keys = Parser.querystring_pattern.findall(method_info.url)
        query_args = {arg: val for arg, val in args.items() if
                      arg in keys and val}

        for key in keys:
            args.pop(key, None)

        url = method_info.url.format(**query_args)

        url_result = urlparse(url)
        if "{" in url_result.path:
            logger.warning(f"Not formatted args in url path: {url}")

        querystring = "&".join(
            f"{k}={v}" for k, v in parse_qsl(url_result.query) if "{" not in v
        )
        return url_result.path + "?" + querystring if querystring else url_result.path

    @staticmethod
    def dict_to_body(func_args: Dict[str, Any]) -> Dict:
        keys = list(func_args.keys())
        if len(keys) == 1:
            return func_args[keys[0]]
        return {}

    @staticmethod
    def get_request(method_info: MethodInfo, *args, **kwargs):
        func_args: Dict[str, Any] = Parser._apply_args(method_info, *args, **kwargs)
        url: str = Parser._get_url(method_info, func_args)

        body = Parser.dict_to_body(func_args)
        request_headers = body.pop("request_headers", None)

        if method_info.form_body:
            data, json = body, {}
        else:
            data, json = {}, body

        return HttpRequest(
            url=url,
            data=data,
            json_body=json,
            method=method_info.http_method,
            request_headers=request_headers
        )


class Container:
    def __init__(self):
        self.protocol_clients = {}
        self.func_bindings = {}

    def bind_protocol(self, protocol, client):
        self.protocol_clients[protocol] = client

    def bind_func(self, func, method_info: MethodInfo):
        self.func_bindings[func] = method_info

    def do_func(self, func, *args, **kwargs):
        method_info: MethodInfo = self.func_bindings.get(func)
        if not method_info:
            raise ValueError(f"No router info found for {func}")

        instance = args[0]
        client: AbstractClient = self.protocol_clients.get(instance)
        request = Parser.get_request(method_info, *args, **kwargs)
        return client.do_request(request)


container = Container()



from pydantic_client.schema.client_config import FactoryConfig


class PydanticClientFactory:
    """
    toml file example:
    [[tools.pydantic_client.factory]]
    name = "book_client
    base_url = "https://example.com/api/v1"
    timeout = 1
    [[tools.pydantic_client.factory]]
    name = "author_client
    base_url = "https://example.com/api/v2"
    timeout = 1
    [[tools.pydantic_client.factory]]
    name = "address_client
    base_url = "https://example.com/api/v3"
    timeout = 1

    factory = PydanticClientFactory.from_toml("pydantic_client.toml") \
        .register_client("book_client", RequestsClient, BookProtocol) \
        .register_client("author_client", HttpxClient, AuthorProtocol)
        .build()

    book: Book = factory.get_client(BookProtocol).get_books(1)

    """

    def __init__(self, config: FactoryConfig):
        # name: client_class
        self.pydantic_clients = {}
        self.config = config

    @classmethod
    def from_toml(cls, toml_file: str) -> "PydanticClientFactory":
        config = FactoryConfig.load_toml(toml_file)
        return cls(config)

    def register_client(
        self,
        client_name: str,
        client_class,
        protocol_class,
        *args,
        **kwargs
    ):
        """
        params:
            client_name: str, name of the client which in toml file.
            client_class: client class which in pydantic-client.
            protocol_class: protocol class which define the routers.
        """
        cfg = self.config.get_by_name(client_name)
        if not cfg:
            raise ValueError(f"client {client_name} not found in config")

        from pydantic_client import PydanticClient

        client = PydanticClient(cfg)

        self.pydantic_clients[protocol_class] = client \
            .bind_client(client_class) \
            .bind_protocol(protocol_class, *args, **kwargs) \
            .build()
        return self

    def build(self):
        return self

    def get_client(self, protocol_class):
        client = self.pydantic_clients.get(protocol_class)
        if not client:
            raise ValueError(f"client for {protocol_class} not found")
        return client



from typing import Type

from pydantic_client.clients.abstract_client import AbstractClient
from pydantic_client.container import container
from pydantic_client.schema.client_config import ClientConfig


class PydanticClient:
    """

    class User(BaseModel):
        id: int
        name: str

    class WebClient:
        @get("/users/{user_id}")
        def get_user(self, user_id: int) -> User:
            ...

    web_client = PydanticClient.from_toml(...)
        .bind_client(RequestsClient)
        .bind_protocol(WebClient)
        .build()

    # or

    client: WebClient = PydanticClient(
        ClientConfig(
            base_url="https://example.com",
            headers={"Authorization": "Bearer abcdefg"},
            timeout=10
        )
    ).bind_client(RequestsClient)
        .bind_protocol(WebClient)
        .build()

    user: User = web_client.get_user(123)

    """

    def __init__(self, config: ClientConfig):
        self.config = config
        self.client = None
        self.protocol = None

    @classmethod
    def from_toml(cls, toml_path: str) -> "PydanticClient":
        config = ClientConfig.load_toml(toml_path)
        return cls(config=config)

    def bind_client(self, client_class: Type[AbstractClient]):
        self.client = client_class(self.config)
        return self

    def bind_protocol(self, protocol_class, *args, **kwargs):
        self.protocol = protocol_class(*args, **kwargs)
        return self

    def build(self):
        container.bind_protocol(self.protocol, self.client)
        return self.protocol



import inspect
import logging
from functools import partial, wraps
from typing import Callable

from pydantic._internal._model_construction import ModelMetaclass

from pydantic_client.container import container
from pydantic_client.schema.method_info import MethodInfo

logger = logging.getLogger(__name__)


def parse_func(
    *,
    url: str,
    func: Callable,
    method: str,
    form_body: bool,
):
    spec = inspect.getfullargspec(func)
    annotations = spec.annotations.copy()
    return MethodInfo(
        func=func,
        http_method=method,
        url=url,
        request_type=annotations,
        response_type=annotations.pop("return", None),
        form_body=form_body
    )


def rest(
    url,
    method: str = "GET",
    form_body: bool = False
) -> Callable:
    if not url.startswith("/"):
        raise ValueError("url must start with slash.")

    def wrapper(func: Callable):
        method_info = parse_func(
            url=url,
            func=func,
            method=method,
            form_body=form_body
        )
        container.bind_func(func, method_info)

        def convert(value, target_type):
            rt = method_info.response_type
            if not rt:
                return value
            if isinstance(rt, ModelMetaclass):
                return target_type(**value)
            try:
                return target_type(value)
            except Exception as e:
                logger.warning(
                    f"Failed to convert {value} to {target_type}: {e}")
                return value

        @wraps(func)
        async def async_wrapped_func(*args, **kwargs):
            ret = container.do_func(func, *args, **kwargs)
            if inspect.isawaitable(ret):
                ret = await ret
            return convert(ret, method_info.response_type)

        @wraps(func)
        def wrapped_func(*args, **kwargs):
            ret = container.do_func(func, *args, **kwargs)
            return convert(ret, method_info.response_type)

        return async_wrapped_func if (
            inspect.iscoroutinefunction(func)) else wrapped_func

    return wrapper


get = partial(rest, method="GET")
post = partial(rest, method="POST")
delete = partial(rest, method="DELETE")
put = partial(rest, method="PUT")
patch = partial(rest, method="PATCH")



from typing import Any, Dict, Tuple, Optional

from pydantic_client.schema.client_config import ClientConfig
from pydantic_client.schema.http_request import HttpRequest


class AbstractClient:

    def __init__(self, config: ClientConfig):
        self.config = config
        self.base_url = config.base_url.rstrip("/")

    def do_request(self, request: HttpRequest) -> Any:
        raise NotImplementedError

    @staticmethod
    def parse_request(request: HttpRequest) -> Tuple[Dict, Dict]:
        if request.data:
            data, json = request.data, None
        else:
            data, json = None, request.json_body
        return data, json

    @classmethod
    def from_toml(cls, toml_config: ClientConfig):
        return cls(toml_config)



from typing import Any

from requests import Session

from pydantic_client.clients.abstract_client import AbstractClient
from pydantic_client.schema.http_request import HttpRequest


class RequestsClient(AbstractClient):
    session = Session()

    def do_request(self, request: HttpRequest) -> Any:
        data, json = self.parse_request(request)
        headers = request.request_headers if request.request_headers \
            else self.config.headers

        try:
            return self.session.request(
                url=self.base_url + request.url,
                method=request.method,
                json=json,
                data=data,
                headers=headers,
                timeout=self.config.timeout,
            ).json()
        except BaseException as e:
            raise e



from typing import Any

from aiohttp.client import ClientSession

from pydantic_client.clients.abstract_client import AbstractClient
from pydantic_client.schema.http_request import HttpRequest


class AIOHttpClient(AbstractClient):

    async def do_request(self, request: HttpRequest) -> Any:
        data, json = self.parse_request(request)
        headers = request.request_headers if request.request_headers \
            else self.config.headers
        async with ClientSession() as session:
            try:
                req = session.request(
                    url=self.base_url + request.url,
                    method=request.method,
                    json=json,
                    data=data,
                    headers=headers,
                    timeout=self.config.timeout
                )

                async with req as resp:
                    resp.raise_for_status()
                    if resp.status == 200:
                        return await resp.json()
            except BaseException as e:
                raise e



import logging
from typing import Dict, Any, Optional

from pydantic import BaseModel

logger = logging.getLogger(__name__)


class ClientConfig(BaseModel):
    base_url: str
    headers: Dict[str, Any] = {}
    http2: bool = False
    timeout: Optional[int] = None

    @classmethod
    def load_toml(cls, path: str):
        try:
            import toml
        except ImportError:
            raise ImportError("toml is required, please install it using pip")

        with open(path) as f:
            config = toml.load(f)
            cfg = config.get("tools", {}) \
                .get("pydantic-client", {}).get("config")
            if not cfg:
                raise ValueError(
                    "`tools.pydantic-client.config` not found in toml file")  # noqa
            return cls(**cfg)


class FactoryConfig(BaseModel):
    config: Dict[str, ClientConfig]

    def get_by_name(self, name: str) -> Optional[ClientConfig]:
        return self.config.get(name)

    @classmethod
    def load_toml(cls, path: str):
        try:
            import toml
        except ImportError:
            raise ImportError("toml is required, please install it using pip")

        with open(path) as f:
            config = toml.load(f)
            cfg = config.get("tools", {}) \
                .get("pydantic-client", {}).get("factory", [])
            if not cfg:
                raise ValueError(
                    "`tools.pydantic-client.factory` not found in toml file")  # noqa

            factory_config = {}
            for config in cfg:
                name = config.pop("name", None)
                if not name:
                    logger.warning("name not found in factory config")
                    continue
                factory_config[name] = ClientConfig(**config)

            return cls(config=factory_config)



from typing import Dict, Optional

from pydantic import BaseModel


class HttpRequest(BaseModel):
    data: Optional[Dict] = {}
    json_body: Optional[Dict] = {}
    url: str
    method: str
    request_headers: Optional[Dict] = None
